import functools
import json
import logging
import math
import os
import re
import time
import traceback
from enum import Enum
from typing import Any, cast

from botocore.exceptions import NoCredentialsError
from datasets import IterableDataset
from evalassist.judges import Criteria
from fastapi import HTTPException
from ibm_watsonx_ai.wml_client_error import (
    ApiRequestFailure,
    CannotSetProjectOrSpace,
    WMLClientError,
)
from openai import AuthenticationError, NotFoundError
from unitxt.inference import (
    CrossProviderInferenceEngine,
    HFAutoModelInferenceEngine,
    InferenceEngine,
    LiteLLMInferenceEngine,
    WMLInferenceEngineChat,
    WMLInferenceEngineGeneration,
)
from unitxt.llm_as_judge import EvaluatorNameEnum, ModelProviderEnum

from .const import CUSTOM_MODELS_PATH, EVAL_ASSIST_DIR
from .extended_unitxt import (
    EXTENDED_EVALUATOR_TO_MODEL_ID,
    EXTENDED_EVALUATORS_METADATA,
    ExtendedEvaluatorMetadata,
    ExtendedEvaluatorNameEnum,
    ExtendedModelProviderEnum,
)
from .judges import Instance

logger = logging.getLogger(__name__)


def fill_unknown_template(template: str, value: str) -> str:
    # Case 1: Unnamed placeholder(s) like "hello {}"
    if "{}" in template:
        return template.format(value)

    # Case 2: Named placeholder like "hello {world}"
    match = re.search(r"\{(\w+)\}", template)
    if match:
        var_name = match.group(1)
        return template.format(**{var_name: value})

    # Case 3: No placeholder
    return template


def get_custom_models():
    if os.path.exists(CUSTOM_MODELS_PATH):
        with open(CUSTOM_MODELS_PATH, "r", encoding="utf-8") as file:
            try:
                custom_models = json.load(file)

                logger.debug(
                    "Loaded the following custom models",
                    json.dumps(custom_models, indent=2),
                )
                return custom_models
            except Exception:
                raise ValueError("${CUSTOM_MODELS_PATH} must be a valid json file")
    else:
        return []


def convert_model_name_wx_to_hf(wx_model_name):
    model_map = {
        "ibm/granite-guardian-3-2b": "ibm-granite/granite-guardian-3.1-2b",
        "ibm/granite-guardian-3-8b": "ibm-granite/granite-guardian-3.1-8b",
    }
    try:
        return model_map[wx_model_name]
    except KeyError:
        return wx_model_name


def get_enum_by_value(value: str, enum: type[Enum]) -> Enum | None:
    for enum_member in enum:
        if enum_member.value == value:
            return enum_member
    return None


def get_local_hf_inference_engine_params(
    model_name: str,
):
    return {
        "model_name": convert_model_name_wx_to_hf(model_name),
        "max_new_tokens": 1024,
        "device": get_default_torch_device(),
    }


def get_cross_inference_engine_params(
    credentials: dict,
    provider: ModelProviderEnum,
    model_name: str,
    custom_params: dict | None = None,
    provider_specific_params: dict | None = None,
):
    provider_specific_args = {}
    inference_engine_params: dict[str, Any] = {
        "max_tokens": 1024,
        "credentials": credentials.copy(),
        "provider": provider,
    }
    if provider == ModelProviderEnum.WATSONX:
        provider_specific_args[ModelProviderEnum.WATSONX.value] = {
            "max_requests_per_second": 7
        }

    if (
        provider == ModelProviderEnum.AZURE
        or provider == ExtendedModelProviderEnum.OPENAI_LIKE
    ):
        inference_engine_params["credentials"]["api_base"] = fill_unknown_template(
            credentials["api_base"], model_name
        )

    is_model_supported_by_cross_inference = (
        provider.value in CrossProviderInferenceEngine.provider_model_map
        and model_name
        in CrossProviderInferenceEngine.provider_model_map[provider.value]
    )

    cross_inference_litellm_keys = [
        provider
        for provider, klass in CrossProviderInferenceEngine._provider_to_base_class.items()
        if klass == LiteLLMInferenceEngine
    ]

    if (
        not is_model_supported_by_cross_inference
        and provider.value in cross_inference_litellm_keys
        and not model_name.startswith(provider.value)
    ):
        # LiteLLM expects the model_name to start with the provider
        # If the user provided a custom model_name that uses litellm, adapt it to start with the provider
        model_name = provider.value + "/" + model_name

    elif provider == ExtendedModelProviderEnum.OPENAI_LIKE:
        provider = ModelProviderEnum.OPENAI

    if custom_params is not None:
        inference_engine_params.update(custom_params)
    if provider_specific_params is not None:
        provider_specific_args.update(provider_specific_params)
    inference_engine_params["model"] = model_name
    inference_engine_params["provider"] = provider.value
    inference_engine_params["provider_specific_args"] = provider_specific_args

    if provider == ExtendedModelProviderEnum.HF_LOCAL:
        if "seed" in inference_engine_params:
            del inference_engine_params["seed"]
        if "credentials" in inference_engine_params:
            del inference_engine_params["credentials"]

    return inference_engine_params


def get_cross_inference_engine(
    credentials: dict[str, str],
    provider: ModelProviderEnum,
    model_name: EvaluatorNameEnum,
    custom_params: dict | None = None,
    provider_specific_params: dict | None = None,
):
    inference_engine_params = get_cross_inference_engine_params(
        credentials=credentials,
        provider=provider,
        model_name=model_name,
        custom_params=custom_params,
        provider_specific_params=provider_specific_params,
    )

    return CrossProviderInferenceEngine(**inference_engine_params)


def get_watsonx_inference_engine(
    credentials: dict[str, str],
    provider: ModelProviderEnum,
    model_name: EvaluatorNameEnum,
    custom_params: dict | None = None,
    use_chat: bool = True,
):
    converted_model_name = "/".join(
        CrossProviderInferenceEngine.provider_model_map["watsonx"][model_name].split(
            "/"
        )[1:]
    )
    inference_engine_params = {
        "max_tokens": 1024,
        "credentials": {
            "api_key": credentials["api_key"],
            "url": credentials["api_base"],
        },
        "model_name": converted_model_name,
    }

    if custom_params is not None:
        inference_engine_params.update(custom_params)

    if "project_id" in credentials and credentials["project_id"] is not None:
        inference_engine_params["credentials"]["project_id"] = credentials["project_id"]
    elif "space_id" in credentials and credentials["space_id"] is not None:
        inference_engine_params["credentials"]["space_id"] = credentials["space_id"]
    else:
        raise ValueError("Must provide either project_id or space_id")

    if "seed" in inference_engine_params and not use_chat:
        inference_engine_params["random_seed"] = inference_engine_params["seed"]
        del inference_engine_params["seed"]

    if not use_chat:
        if "seed" in inference_engine_params:
            inference_engine_params["random_seed"] = inference_engine_params["seed"]
            del inference_engine_params["seed"]
        if "max_tokens" in inference_engine_params:
            inference_engine_params["max_new_tokens"] = inference_engine_params[
                "max_tokens"
            ]
            del inference_engine_params["max_tokens"]

    return (
        WMLInferenceEngineChat(**inference_engine_params)
        if use_chat
        else WMLInferenceEngineGeneration(**inference_engine_params)
    )


preloaded_hf_models = {}


def get_hf_inference_engine(
    model_name: str,
    custom_params: dict | None = None,
):
    global preloaded_hf_models
    if model_name in preloaded_hf_models:
        logger.debug(f"Using preloaded HF model {model_name}")
        return preloaded_hf_models[model_name]
    else:
        logger.debug(f"Loading model {model_name}")
        params = get_local_hf_inference_engine_params(model_name)
        if custom_params is not None:
            params.update(custom_params)
        if "seed" in params:
            del params["seed"]
        hf_model = HFAutoModelInferenceEngine(**params)
        preloaded_hf_models[model_name] = hf_model
        return hf_model


def get_inference_engine(
    credentials: dict[str, str | None],
    provider: ModelProviderEnum | ExtendedModelProviderEnum,
    model_name: str,
    custom_params: dict | None = None,
    provider_specific_params: dict | None = None,
) -> InferenceEngine:
    # if provider == ExtendedModelProviderEnum.LOCAL_HF:
    #     return get_hf_inference_engine(model_name, custom_params)
    if provider == ModelProviderEnum.WATSONX and (
        "granite-guardian" in model_name
        or ("space_id" in credentials and credentials["space_id"])
    ):
        use_chat = "granite-guardian" not in model_name  # uses probs
        return get_watsonx_inference_engine(
            credentials, provider, model_name, custom_params, use_chat
        )
    return get_cross_inference_engine(
        credentials, provider, model_name, custom_params, provider_specific_params
    )


def get_model_name_from_evaluator(
    evaluator_metadata: ExtendedEvaluatorMetadata,
    provider: str,
) -> str:
    model_name = EXTENDED_EVALUATOR_TO_MODEL_ID.get(evaluator_metadata.name, None)
    return (
        evaluator_metadata.custom_model_path
        if evaluator_metadata.custom_model_path is not None
        else model_name
    )


def get_enum_values(e: type[Enum]):
    return [member.value for member in e]


def get_evaluator_metadata_wrapper(
    evaluator_name: EvaluatorNameEnum | ExtendedEvaluatorNameEnum,
    custom_model_name: str | None = None,
) -> ExtendedEvaluatorMetadata:
    if evaluator_name.name != ExtendedEvaluatorNameEnum.CUSTOM.name:
        evaluator_search = [
            e
            for e in EXTENDED_EVALUATORS_METADATA
            if e.name.name == evaluator_name.name
        ]
        if len(evaluator_search) == 0:
            raise ValueError(f"An evaluator with id {evaluator_name} does not exist.")
        if len(evaluator_search) > 1:
            raise ValueError(
                f"An evaluator with id {evaluator_name} matched several models."
            )
        return evaluator_search[0]
    else:
        custom_models = get_custom_models()
        if custom_model_name not in [
            custom_model["name"] for custom_model in custom_models
        ]:
            raise ValueError("The specified custom model was not found")

        custom_model = [
            custom_model
            for custom_model in custom_models
            if custom_model["name"] == custom_model_name
        ][0]
        providers: list[ModelProviderEnum | ExtendedModelProviderEnum] = []
        for p in providers:
            try:
                providers.append(
                    ExtendedModelProviderEnum(p)
                    if get_enum_values(ExtendedModelProviderEnum)
                    else ModelProviderEnum(p)
                )
            except Exception as e:
                print(e)

        return ExtendedEvaluatorMetadata(
            name=evaluator_name,
            custom_model_name=custom_model["name"],
            custom_model_path=custom_model["path"],
            providers=providers,
        )


def get_default_torch_device(avoid_mps: bool = False):
    try:
        from unitxt.inference import TorchDeviceMixin

        device = TorchDeviceMixin().get_device_id()
        return device
    except Exception:
        raise ImportError(
            'torch is not installed -by default-. Make sure you installed local inference dependencies with pip install "evalassist[local-inference]"'
        )
    logger.debug(f"Detected device: {device}")


def init_evaluator_name(
    evaluator_name: EvaluatorNameEnum | ExtendedEvaluatorNameEnum | str,
) -> tuple[ExtendedEvaluatorNameEnum, str | None]:
    evaluator_name_as_enum = get_enum_by_value(evaluator_name, EvaluatorNameEnum)
    if evaluator_name_as_enum is None:
        evaluator_name_as_enum = get_enum_by_value(
            evaluator_name, ExtendedEvaluatorNameEnum
        )

    if evaluator_name_as_enum is not None:
        return (evaluator_name_as_enum, None)

    custom_model_name = evaluator_name
    evaluator_name = ExtendedEvaluatorNameEnum.CUSTOM
    return (evaluator_name, custom_model_name)


def log_runtime(function):
    """
    Usage: wrap a function call with the log_runtime function to log its runtime
    """

    @functools.wraps(function)
    def wrapper(*args, **kwargs):
        start_time = time.time()

        res = function(*args, **kwargs)

        end_time = time.time()
        total_time = round(end_time - start_time, 2)

        logger.debug(
            f"{function.__name__} took {total_time} seconds, {round(total_time / 60, 2)} minutes"
        )

        return res

    return wrapper


def to_snake_case(name: str) -> str:
    return name.replace(" ", "_").lower()


def parse_exception_message(e: Exception) -> str:
    error_message = e.message if hasattr(e, "message") else None
    if error_message is None:
        return str(e)
    result = error_message.split(" - ", 1)
    if len(result) == 1:
        return result[0]
    try:
        result = json.loads(result[1])
    except json.JSONDecodeError:
        return result[1]
    if "errorMessage" in result:
        return result["errorMessage"]
    elif (
        "errors" in result
        and len(result["errors"]) > 0
        and "message" in result["errors"][0]
    ):
        return result["errors"][0]["message"]
    elif "title" in result:
        return result["detail"]
    else:
        return str(result)


def handle_exception(e: Exception) -> None:
    error_message = parse_exception_message(e)
    if isinstance(e, NotFoundError):
        raise HTTPException(status_code=400, detail=error_message)
    elif isinstance(e, ValueError):
        raise HTTPException(status_code=400, detail=error_message)
    elif isinstance(e, ApiRequestFailure):
        raise HTTPException(status_code=400, detail=error_message)
    elif isinstance(e, NoCredentialsError):
        raise HTTPException(
            status_code=400,
            detail=error_message,
        )
    elif isinstance(e, AuthenticationError):
        raise HTTPException(status_code=400, detail=error_message)
    elif isinstance(e, CannotSetProjectOrSpace):
        raise HTTPException(status_code=400, detail=error_message)
    elif isinstance(e, WMLClientError):
        raise HTTPException(status_code=400, detail=error_message)
    elif isinstance(e, AssertionError):
        raise HTTPException(status_code=400, detail=error_message)
    elif isinstance(e, RuntimeError):
        if e.__cause__:
            handle_exception(e.__cause__)
        else:
            raise HTTPException(400, error_message)
    else:
        raise HTTPException(
            status_code=400,
            detail=error_message,
        )


def handle_llm_generation_exceptions(func):
    def wrapper(*args, **kwargs):
        try:
            return func(*args, **kwargs)
        except Exception as e:
            handle_exception(e)
        finally:
            traceback.print_exc()

    return wrapper


def clean_object(results: dict | list):
    """Cleans the object by removing `None` values and empty lists and dictionaries.

    Args:
        results (Union[dict, list]): The results to clean.

    Returns:
        Union[dict, list]: The cleaned results.
    """
    if isinstance(results, list):
        cleaned_list = [
            clean_object(x)
            for x in results
            if x is not None and not (isinstance(x, (list, dict)) and len(x) == 0)
        ]
        return cleaned_list
    elif isinstance(results, dict):
        cleaned = {
            k: (v if not isinstance(v, dict) else clean_object(v))
            for k, v in results.items()
            if v is not None and not (isinstance(v, (list, dict)) and len(v) == 0)
        }
        # Remove the dictionary itself if it becomes empty
        return {
            k: v
            for k, v in cleaned.items()
            if not (isinstance(v, dict) and len(v) == 0)
        }
    else:
        return results


def get_system_version():
    try:
        # git is a dev dependency so import may fail
        try:
            import git
        except ImportError as e:
            logger.error(
                "Make sure you installed EvalAssist's dev dependencies with poetry install --with dev"
            )
            raise e
        repo = git.repo.Repo(EVAL_ASSIST_DIR.parent.parent.parent)
        version = repo.git.describe(tags=True)
        source = "git"
    except Exception:
        try:
            from importlib.metadata import version

            version = version("evalassist")
            source = "pypi"

            if version == "0.0.0":
                # version is 0.0.0 in the toml only when
                # evalassist is executed using poetry
                # and the .git folder is not present in the root directory
                raise ValueError("Invalid version")
        except Exception as e:
            version = None
            source = None
            print(e)
            logging.warning("Could not get EvalAssist version")

    return {"version": version, "source": source}


def unitxt_dataset_to_evalassist_instances(
    dataset: IterableDataset,
    criteria: list[Criteria],
) -> list[Instance]:
    if any(
        criterion.to_evaluate_field is None or criterion.context_fields is None
        for criterion in criteria
    ):
        raise ValueError(
            "The criteria.to_evaluate_field is None. It must be set to retrieve the response to evaluate from the task_data"
        )
    task_data_list = [json.loads(d["task_data"]) for d in dataset]
    return [
        Instance(
            fields={
                **{k: task_data[k] for k in cast(list, criterion.context_fields)},
                criterion.to_evaluate_field: task_data[criterion.to_evaluate_field],
            },
        )
        for task_data, criterion in zip(task_data_list, criteria)
    ]


def get_inference_engine_from_judge_metadata(
    evaluator_name: ExtendedEvaluatorNameEnum,
    custom_model_name: str | None,
    provider: ModelProviderEnum | ExtendedModelProviderEnum,
    llm_provider_credentials: dict[str, str | None],
    custom_params: dict[str, Any] | None = None,
):
    """
    Get an inference engine from evaluator's data.
    """
    evaluator_metadata = get_evaluator_metadata_wrapper(
        evaluator_name,
        custom_model_name,
    )
    model_name = get_model_name_from_evaluator(
        evaluator_metadata,
        provider,
    )
    return get_inference_engine(
        credentials=llm_provider_credentials,
        provider=provider,
        model_name=model_name,
        custom_params=custom_params,
    )


def convert_nan_to_none(obj):
    if isinstance(obj, float) and math.isnan(obj):
        return None
    elif isinstance(obj, dict):
        return {k: convert_nan_to_none(v) for k, v in obj.items()}
    elif isinstance(obj, list):
        return [convert_nan_to_none(v) for v in obj]
    return obj



import logging
from typing import cast

from ..base import BaseDirectJudge, UnitxtInferenceEngineMixin
from ..direct_judge import DirectJudge
from ..types import Criteria, CriteriaOption, DirectInstanceResult, Instance

logger = logging.getLogger(__name__)


class CriticizedDirectJudge(BaseDirectJudge, UnitxtInferenceEngineMixin):
    def get_name(self) -> str:
        return "criticized"

    def _run(
        self,
        instances: list[Instance],
        criteria: list[Criteria],
    ) -> list[DirectInstanceResult]:
        simple_judge = DirectJudge(self.inference_engine)
        logger.info(f"Running evaluation on {len(instances)} instances...")
        results = simple_judge(instances, criteria)
        criticize_criterion = Criteria(
            name="Judge quality",
            description=(
                "Evaluate the quality of the judge’s reasoning and decision. "
                "Consider whether the explanation is logically consistent, well-grounded "
                "in the provided judge prompt, and whether the selected option is justified. "
            ),
            context_fields=["judge_prompt", "explanation"],
            to_evaluate_field="selected_option",
            options=[
                CriteriaOption(
                    name="Excellent",
                    description="The reasoning is clear, well-structured, and fully aligned with the judge prompt. The chosen option is correct and well-justified.",
                    score=1.0,
                ),
                CriteriaOption(
                    name="Acceptable",
                    description="The reasoning and choice are generally sound, but there may be minor issues in clarity, completeness, or justification.",
                    score=0.5,
                ),
                CriteriaOption(
                    name="Could be Improved",
                    description="The reasoning or chosen option shows noticeable gaps, inconsistencies, or unclear justification. Feedback is needed to help improve the judge.",
                    score=0.25,
                ),
                CriteriaOption(
                    name="Bad",
                    description="The reasoning is flawed, unclear, or inconsistent with the judge prompt, and the chosen option is incorrect or poorly justified.",
                    score=0.0,
                ),
            ],
        )

        criticize_judge = DirectJudge(
            self.inference_engine,
            generate_feedback=True,
        )

        logger.info("Criticizing the results...")
        criticized_results = criticize_judge(
            instances=[
                Instance(
                    fields={
                        "judge_prompt": str(result.metadata["prompt"]),
                        "explanation": result.explanation,
                        "selected_option": str(result.selected_option),
                    },
                )
                for result in results
            ],
            criteria=criticize_criterion,
        )

        to_revisit_results_index = [
            i
            for i, criticized_result in enumerate(criticized_results)
            if cast(float, criticized_result.score) < 0.5
        ]
        logger.info(
            f"Running the evaluation again taking the feedback into account for {len(to_revisit_results_index)} instances..."
        )
        revisited_results = {}
        if len(to_revisit_results_index) > 0:
            revisited_results = {
                to_revisit_result_index: revisited_result
                for to_revisit_result_index, revisited_result in zip(
                    to_revisit_results_index,
                    simple_judge(
                        instances=[
                            Instance(
                                fields={
                                    **instances[to_revisit_result_index].fields,
                                    "Feedback from previous evaluation": f"The previous time you evaluated this instance, your evaluation was categorized as '{criticized_results[to_revisit_result_index].selected_option}'. In order to improve it, take the following feedback into account: {cast(str, criticized_results[to_revisit_result_index].feedback)}"
                                    if criticized_results[
                                        to_revisit_result_index
                                    ].feedback
                                    != ""
                                    else "Nothing to improve",
                                },
                            )
                            for to_revisit_result_index in to_revisit_results_index
                        ],
                        criteria=[
                            Criteria(
                                description=f"{criteria[to_revisit_result_index].description} Important: take the judge feedback into account to improve your judgement",
                                name=criteria[to_revisit_result_index].name,
                                options=criteria[to_revisit_result_index].options,
                                context_fields=[
                                    *(
                                        cast(
                                            list,
                                            criteria[
                                                to_revisit_result_index
                                            ].context_fields,
                                        )
                                        if criteria[
                                            to_revisit_result_index
                                        ].context_fields
                                        is not None
                                        else []
                                    ),
                                    "judge_feedback",
                                ],
                                to_evaluate_field=criteria[
                                    to_revisit_result_index
                                ].to_evaluate_field,
                            )
                            for to_revisit_result_index in to_revisit_results_index
                        ],
                    ),
                )
            }

        for result in results:
            if result.metadata is None:
                result.metadata = {}
            result.metadata["original_result"] = result.model_dump()
            result.metadata["required_improvement"] = False

        for to_revisit_result_index, revisited_result in revisited_results.items():
            if revisited_result.metadata is None:
                revisited_result.metadata = {}
            revisited_result.metadata["required_improvement"] = True
            revisited_result.metadata["original_result"] = results[
                to_revisit_result_index
            ].model_dump()

        final_results = [
            revisited_results[i] if i in revisited_results else results[i]
            for i in range(len(instances))
        ]

        for result, criticized_result in zip(final_results, criticized_results):
            if result.metadata is None:
                result.metadata = {}
            result.metadata["meta_judge_result"] = criticized_result.model_dump()

        return final_results



# import json
# import logging

# from unitxt.inference import CrossProviderInferenceEngine

# from .base import BaseDirectJudge, Criteria, DirectInstance, DirectInstanceResult

# logger = logging.getLogger(__name__)


# def get_risk_name(unparsed_risk_name: str):
#     risk_name = unparsed_risk_name
#     risk_name = (
#         "_".join(risk_name.split(" ")[(1 if risk_name.startswith("Context") else 2) :])
#         .lower()
#         .replace(" ", "_")
#     )

#     return risk_name if risk_name != "general_harm" else "harm"


# field_map = {
#     "user_message_field": "user_message",
#     "assistant_message_field": "assistant_message",
#     "context_field": "context",
#     "tools_field": "tools",
# }

# criteria_ids = [
#     "social_bias",
#     "jailbreak",
#     "profanity",
#     "sexual_content",
#     "unethical_behavior",
#     "violence",
#     "harm",
#     "groundedness",
#     "answer_relevance",
#     "context_relevance",
#     "function_call",
#     "evasiveness",
#     "harm_engagement",
# ]


# class GraniteGuardianJudge(BaseDirectJudge):
#     def get_prompt(self, messages, guardian_config, think: bool = False) -> str:
#         """Todo"""
#         logger.debug(f"Messages are: {json.dumps(messages, indent=2)}")

#         logger.debug(f"guardian_config is: {guardian_config}")
#         prompt = tokenizer.apply_chat_template(
#             messages,
#             guardian_config=guardian_config,
#             tokenize=False,
#             add_generation_prompt=True,
#             think=think,
#         )
#         logger.debug(f"Prompt is:\n{prompt}")
#         return prompt

#     def _run(
#         self,
#         instances: list[DirectInstance],
#         criteria: list[Criteria],
#     ) -> list[DirectInstanceResult]:
#         inference_engine = CrossProviderInferenceEngine(
#             model="granite-guardian-3-3-8b",
#             provider="rits",
#         )
#         risk_names = [get_risk_name(criterion.name) for criterion in criteria]

#         # Custom criteria

#         # set custom_scoring_schema for custom scoring schema



import logging
from collections.abc import Callable
from textwrap import dedent
from typing import Any, Literal, cast

from evalassist.judges.batch_repair_parser import BatchRepairParser
from evalassist.judges.utils import (
    build_format_instructions,
    generate_dynamic_pydantic_model,
    get_context_dict,
    get_to_evaluate_text,
    is_float,
)
from pydantic import BaseModel, Field
from unitxt.inference import InferenceEngine

from .base import BaseDirectJudge, JudgeDescriptor, UnitxtInferenceEngineMixin
from .types import Criteria, DirectInstanceResult, Instance

logger = logging.getLogger(__name__)


class DirectJudge(BaseDirectJudge, UnitxtInferenceEngineMixin):
    generate_synthetic_persona: bool
    generate_feedback: bool
    judge_description_prompt: str | None
    on_generation_failure: Literal["raise", "random"]

    def __init__(
        self,
        inference_engine: InferenceEngine,
        generate_synthetic_persona: bool = False,
        judge_description_prompt: str | None = None,
        generate_feedback: bool = False,
        self_consistency: bool | int = False,
        on_generation_failure: Literal["raise", "random"] = "random",
        *args,
        **kwargs,
    ):
        super().__init__(
            inference_engine=inference_engine,
            self_consistency=self_consistency,
            *args,
            **kwargs,
        )

        if on_generation_failure not in ["raise", "random"]:
            raise ValueError(
                "on_generation_failure must be either 'raise' or 'random'. Received {on_generation_failure} instead."
            )
        self.on_generation_failure = on_generation_failure

        if generate_synthetic_persona and judge_description_prompt:
            raise ValueError(
                "Either provide set generate_synthetic_persona to False or don't provide a judge_description_prompt."
            )

        if self.self_consistency is True or self.self_consistency > 1:
            temp = getattr(self.inference_engine, "temperature", None)
            if temp is not None:
                try:
                    if float(temp) == 0.0:
                        logger.warning(
                            "Self-consistency may not bring any benefit when temperature is 0."
                        )
                except (TypeError, ValueError):
                    logger.debug(
                        "Could not interpret temperature value for self-consistency check."
                    )

        self.generate_synthetic_persona = generate_synthetic_persona
        self.judge_description_prompt = judge_description_prompt
        self.generate_feedback = generate_feedback

    def get_name(self) -> str:
        return f"in_house{'_with_synthetic_persona' if self.generate_synthetic_persona else ''}{'_with_feedback' if self.generate_feedback else ''}{f'_with_self_consistency_{self.self_consistency}_attempts' if self.self_consistency else ''}"

    def get_descriptor(self) -> JudgeDescriptor:
        judge_descriptor = JudgeDescriptor(self.get_name(), "direct", "")
        judge_descriptor.inference_engine_id = self.get_inference_engine_id()
        return judge_descriptor

    def generate_personas(
        self,
        context_sections: list[str],
        predictions: list[str],
        criteria: list[Criteria],
    ) -> list[tuple[str, str]]:
        unique_criteria_instance: list[tuple[Criteria, tuple[str, str]]] = list(
            {
                criterion.name: (criterion, (context_section, prediction))
                for criterion, context_section, prediction in zip(
                    criteria, context_sections, predictions
                )
            }.values()
        )
        unique_criteria, instance_examples = zip(*unique_criteria_instance)  # type: ignore
        unique_criteria: list[Criteria] = list(unique_criteria)
        instance_examples: list[tuple[str, str]] = list(instance_examples)

        instance_examples_str = [
            context_section + "\nText to evaluate: " + prediction
            for context_section, prediction in instance_examples
        ]

        synthetic_persona_klasses = []
        format_instructions_list = []
        for criterion in unique_criteria:
            dynamic_model = generate_dynamic_pydantic_model(
                model_name="structured_output_model",
                field_definitions=[
                    (
                        "persona_name",
                        str,
                        Field(
                            ...,
                            description=f"The name of the persona responsible for evaluating the {criterion.to_evaluate_field} according to the criterion {criterion.name}.",
                        ),
                        [],
                    ),
                    (
                        "persona_description",
                        str,
                        Field(
                            ...,
                            description="The description of why the <persona_name> is ideal to perform the evaluation. Don't include the the initial 'you'. For example: 'an expert on evaluating text based on a rubric' or 'a customer support specialist experienced in clarity and tone explanation'.",
                        ),
                        [],
                    ),
                ],
            )
            synthetic_persona_klasses.append(dynamic_model)
            format_instructions_list.append(build_format_instructions(dynamic_model))

        template = dedent(
            text="""\
                Your task is to generate a persona that is the most appropriate to evaluate a text based on the following criteria.
                You will be provided with the criteria name, description and options and an example instance.

                ### Criterion:

                {criteria_name_section}
                Description: {criteria_description}
                Options:
                {criteria_options}

                ### Example instance

                {instance_example}

                For the persona, you will generate the name or role (e.g. a doctor, a philosopher, a lawyer) and a brief description that makes emphasis on what makes the persona the ideal for performing the evaluation (e.g. have a lot of experience reading and writing email summaries).

                ### Output format

                The persona info will be used as this:
                "You are <persona_name>. Your task is to evaluate a {to_evaluate_field}. You <persona_description>".

                {format_instruction}
            """
        )

        prompts = [
            template.format(
                criteria_name_section=f"Criteria name: {criterion.name}"
                if criterion.name
                else "",
                criteria_description=criterion.description,
                criteria_options="\n".join(
                    [
                        f"- {o.name}{f': {o.description}' if o.description else ''}"
                        for o in criterion.options
                    ]
                ),
                to_evaluate_field=criterion.to_evaluate_field
                if criterion.to_evaluate_field
                else "text",
                instance_example=instance_example_str,
                format_instruction=format_instruction,
            )
            for criterion, instance_example_str, format_instruction in zip(
                criteria, instance_examples_str, format_instructions_list
            )
        ]

        unparsed_responses: list[str] = cast(
            list[str],
            self.inference_engine.infer(
                dataset=[
                    {"source": prompt, "data_classification_policy": ["public"]}
                    for prompt in prompts
                ]
            ),
        )

        parser = BatchRepairParser(
            inference_engine=self.inference_engine,
            max_retries=3,
            on_generation_failure=self.on_generation_failure,
        )

        parsed_responses, _ = parser.parse_and_repair(
            unparsed_responses=unparsed_responses,
            on_failure_default={"persona_name": "", "persona_description": ""},
            model_classes=synthetic_persona_klasses,
        )

        personas = [
            (persona.persona_name, persona.persona_description)  # type: ignore
            for persona in parsed_responses
        ]  # type: ignore
        criteria_name_to_persona = {
            criterion.name: persona
            for criterion, persona in zip(unique_criteria, personas)
        }
        personas_completed = [
            criteria_name_to_persona[criterion.name] for criterion in criteria
        ]
        return personas_completed

    def generate_pydantic_model(
        self,
        model_name: str,
        criterion: Criteria,
        include_feedback: bool,
    ) -> type[BaseModel]:
        criteria_option_names = [option.name for option in criterion.options]

        def validate_selected_option(cls, value: str) -> str:
            if value not in criteria_option_names:
                raise ValueError(f"value must be one of {criteria_option_names}")
            return value

        field_defs = [
            (
                "explanation",
                str,
                Field(..., description="Step by step explanation of the evaluation"),
                [],
            ),
            (
                "selected_option",
                str,
                Field(
                    ...,
                    description=f"The chosen option. Any of {', '.join(criteria_option_names)}",
                ),
                [validate_selected_option],
            ),
        ]
        model: type[BaseModel]
        if not include_feedback:
            model = generate_dynamic_pydantic_model(model_name, field_defs)
        else:
            field_defs.append(
                (
                    "feedback",
                    str,
                    Field(
                        description=f"Actionable suggestions that would help improve the evaluated {criterion.to_evaluate_field if criterion.to_evaluate_field is not None else 'response'} based on the explanation",
                    ),
                    [],
                )
            )
            model = generate_dynamic_pydantic_model(
                f"{model_name}WithFeedback", field_defs
            )
        return model

    def get_in_context_example_as_str(self, criterion: Criteria):
        if not criterion.examples:
            return ""

        title = "\n\n## Examples\n\nTake into account the following examples when performing the evaluation.\n\n"
        examples_str = []
        for i, example in enumerate(criterion.examples):
            context: dict[str, str] = get_context_dict(
                cast(Instance, example.instance), criterion
            )
            context_str: str = (
                "\n\n".join(f"- {k}: {v}" for k, v in context.items())
                if len(context)
                else ""
            )
            context_section_str: str = (
                ("\n#### Context\n\n" + context_str) if context_str else ""
            )

            prediction_section = f"#### The {criterion.to_evaluate_field if criterion.to_evaluate_field else 'text'} to evaluate\n{get_to_evaluate_text(cast(Instance, example.instance), criterion)}"

            explanation_section = (
                f"#### Explanation: {example.explanation}\n\n"
                if example.explanation
                else ""
            )

            ground_truth_section = f"#### Selected option: {example.selected_option}"

            example_str = f"### Example {i + 1}:\n{context_section_str}\n\n{prediction_section}\n\n{explanation_section}{ground_truth_section}\n"
            examples_str.append(example_str)
        res = title + "\n\n".join(examples_str) + "\n[End of examples]"
        return res

    def _run(
        self,
        instances: list[Instance],
        criteria: list[Criteria],
    ) -> list[DirectInstanceResult]:
        format_instructions_list = []
        criteria_options_list = []
        criteria_option_names_list = []
        model_classes = []
        feedback_step_sections = []
        to_evaluate_fields = []
        for criterion in criteria:
            klass = self.generate_pydantic_model(
                model_name=f"{criterion.name}_model",
                criterion=criterion,
                include_feedback=self.generate_feedback,
            )
            model_classes.append(klass)

            criteria_options: str = "\n".join(
                [
                    f"- {option.name}{f': {option.description}' if option.description else ''}"
                    for option in criterion.options
                ]
            )
            criteria_option_names_list.append(
                ", ".join([f'"{option.name}"' for option in criterion.options])
            )

            criteria_options_list.append(criteria_options)

            to_evaluate_field = (
                criterion.to_evaluate_field
                if criterion.to_evaluate_field is not None
                else "response"
            )
            to_evaluate_fields.append(to_evaluate_field)

            feedback_step_section = (
                f'7. At the end, provide "feedback" consisting of actionable suggestions that would help improve the evaluated {to_evaluate_field}. Unlike the explanation, which explains the reasoning behind the judgment, the feedback should focus on guiding refinement. For example, in creative writing, it could suggest improving clarity, coherence, or narrative flow. In analytical tasks, it could recommend strengthening evidence, refining arguments, or correcting inaccuracies. Keep feedback concise and specific enough to support iterative improvement. If you consider that the {to_evaluate_field} is optimal, leave the "feedback" field empty ("")'
                if self.generate_feedback
                else ""
            )
            feedback_step_sections.append(feedback_step_section)
            format_instructions_list.append(build_format_instructions(klass))

        predictions: list[str] = [
            cast(str, get_to_evaluate_text(instance, criterion))
            for instance, criterion in zip(instances, criteria)
        ]
        context_variables_list: list[dict[str, str]] = [
            get_context_dict(instance, criterion)
            for instance, criterion in zip(instances, criteria)
        ]
        str_context_variables_list: list[str | None] = [
            "\n\n".join(f"- {k}: {v}" for k, v in c.items()) if len(c) else None
            for c in context_variables_list
        ]

        context_sections: list[str] = [
            ("\n\n## Context\n\n" + c + "\n") if c is not None else ""
            for c in str_context_variables_list
        ]
        if self.judge_description_prompt:
            judge_description_sections = [self.judge_description_prompt] * len(criteria)
        elif self.generate_synthetic_persona:
            personas = self.generate_personas(
                context_sections=context_sections,
                predictions=predictions,
                criteria=criteria,
            )
            judge_description_sections = [
                f"You are {persona_name}. You are {persona_description}"
                for persona_name, persona_description in personas
            ]
        else:
            judge_description_sections = [
                "You are an expert evaluator. Your task is to objectively and concisely assess a text against a given criterion and optional context. Focus on the substance of the text, follow the instructions carefully, and produce output strictly in the required JSON format. Do not include any explanations, commentary, or text outside the JSON object."
            ] * len(criteria)

        system_template = dedent(
            """\
            {judge_description_section}

            You must output only valid JSON with no extra text.
            Use the following schema and formatting rules:
            {format_instructions}
            """
        )

        user_template = dedent(
            """\
                You will be given:
                - **Criterion** (name, description, options)
                - **Optional context**
                - **The {to_evaluate_field}** to evaluate

                ## Required evaluation behavior (follow these precisely):

                1. Read the *criterion* and the *context* carefully.

                2. Compare the {to_evaluate_field} to the criterion and the context.

                3. Decide which criterion option best fits the {to_evaluate_field}.

                • If more than one option seems plausible:
                    - Choose the one that best satisfies the criterion based on the strongest evidence, not speculation.
                    - If still unclear, select the option that is closest but explain the ambiguity in "explanation".

                4. Write your reasoning in the "explanation", using clear markdown bullet points that describe why one option fits best. When possible, cite specific excerpts or features from the {to_evaluate_field} that support your choice.

                5. Set `"selected_option"` to exactly one of the following values: {criteria_option_names}.

                6. Before submitting your final answer, verify that:
                - The output is valid JSON
                - All keys are quoted
                - There are no trailing commas
                - `"selected_option"` matches exactly one allowed value
                {feedback_step_section}

                ## Criteria:{criteria_name_section}
                Description: {criteria_description}
                Options:
                {criteria_options}{examples_section}{context_section}

                ## The {to_evaluate_field} to evaluate

                {text_to_evaluate}
                """
        )

        prompts: list[list[dict]] = [
            [
                {
                    "role": "system",
                    "content": system_template.format(
                        judge_description_section=judge_description_section,
                        format_instructions=format_instructions,
                    ),
                },
                {
                    "role": "user",
                    "content": user_template.format(
                        text_to_evaluate=prediction,
                        context_section=context_section,
                        examples_section=self.get_in_context_example_as_str(criterion),
                        criteria_name_section=(
                            f"\n\nCriteria name: {criterion.name}"
                            if criterion.name
                            else "\n"
                        ),
                        criteria_description=criterion.description,
                        criteria_options=criterion_options,
                        to_evaluate_field=to_evaluate_field,
                        feedback_step_section=feedback_step_section,
                        criteria_option_names=criteria_option_names,
                    ),
                },
            ]
            for prediction, context_section, criterion, criterion_options, format_instructions, to_evaluate_field, feedback_step_section, judge_description_section, criteria_option_names in zip(
                predictions,
                context_sections,
                criteria,
                criteria_options_list,
                format_instructions_list,
                to_evaluate_fields,
                feedback_step_sections,
                judge_description_sections,
                criteria_option_names_list,
            )
        ]

        dataset = [
            {
                "source": messages,
                "data_classification_policy": ["public"],
            }
            for messages in prompts
        ]

        unparsed_responses: list[str] = cast(
            list[str],
            self.inference_engine(dataset=dataset),
        )

        parser = BatchRepairParser(
            inference_engine=self.inference_engine,
            max_retries=3,
            on_generation_failure=self.on_generation_failure,
        )

        parsed_responses, parsing_metadatas = parser.parse_and_repair(
            unparsed_responses=unparsed_responses,
            on_failure_default=[
                {"selected_option": [option.name for option in criterion.options]}
                for criterion in criteria
            ],
            model_classes=model_classes,
        )
        explanations: list[str] = [r.explanation for r in parsed_responses]  # type: ignore
        selected_options: list[str] = [r.selected_option for r in parsed_responses]  # type: ignore
        feedbacks: list[str | None] = [
            None if not self.generate_feedback else r.feedback  # type: ignore
            for r in parsed_responses  # type: ignore
        ]
        return [
            DirectInstanceResult(
                instance=instance,
                criteria=criterion,
                selected_option=selected_option,
                explanation=explanation,
                feedback=feedback,
                # score=next(iter(option.name for option in criterion.options if option.name == selected_option)).score,
                positional_bias=None,
                metadata={
                    **parsing_metadata,
                    "prompt": prompt,
                    "unparsed_response": unparsed_response,
                },
            )
            for selected_option, explanation, feedback, prompt, unparsed_response, criterion, parsing_metadata, instance in zip(
                selected_options,
                explanations,
                feedbacks,
                prompts,
                unparsed_responses,
                criteria,
                parsing_metadatas,
                instances,
            )
        ]

    def evaluate_with_custom_prompt(
        self,
        judge_prompts: list[str],
        valid_outputs: list[str] | tuple[int, int] | None = None,
    ) -> list[DirectInstanceResult]:
        field_defs: list[tuple[str, type, Any, list[Callable[..., Any]]]] = [
            (
                "explanation",
                str,
                Field(..., description="Step by step explanation of the evaluation"),
                [],
            ),
        ]
        selected_field: Literal["selected_option", "selected_score"]

        if valid_outputs is not None:
            if isinstance(valid_outputs, list):

                def validate_selected_option(cls, value: str) -> str:
                    if value not in valid_outputs:
                        raise ValueError(f"value must be one of {valid_outputs}")
                    return value

                selected_field = "selected_option"
                field_defs.append(
                    (
                        selected_field,
                        str,
                        Field(
                            ...,
                            description=f"The chosen option. Any of {', '.join(valid_outputs)}",
                        ),
                        [validate_selected_option],
                    )
                )
            else:
                if len(valid_outputs) != 2:
                    raise ValueError(
                        "If a tuple is provided as valid_outputs, it must have two element as it is interpreted as a numerical interval."
                    )
                if not isinstance(valid_outputs, tuple):
                    raise ValueError(
                        f"valid_outputs must be of type tuple. Instead, got type {type(valid_outputs)}"
                    )
                if not isinstance(valid_outputs[0], int) or not isinstance(
                    valid_outputs[0], int
                ):
                    raise ValueError(
                        f"valid_outputs's numerical interval got unexpected types, you provided Tuple[{type(valid_outputs[0])}, {type(valid_outputs[1])}]"
                    )

                def validate_selected_score(cls, value: float) -> float:
                    if not (valid_outputs[0] <= value <= valid_outputs[1]):
                        raise ValueError(
                            f"Value must be greater or equal than {valid_outputs[0]} and less or equal than {valid_outputs[1]}"
                        )  # type: ignore
                    return value

                selected_field = "selected_score"
                field_defs.append(
                    (
                        selected_field,
                        int,
                        Field(
                            ...,
                            description=f"The chosen score. A number between {valid_outputs[0]} and {valid_outputs[1]}",
                        ),
                        [validate_selected_score],
                    )
                )
        else:
            selected_field = "selected_score"
            field_defs.append(
                (
                    selected_field,
                    int,
                    Field(
                        ...,
                        description="The chosen option.",
                    ),
                    [],
                )
            )

        dynamic_model = generate_dynamic_pydantic_model(
            "structured_output_model", field_defs
        )

        prompt_template = dedent(
            text="""\
                {judge_prompt}

                ### Output format
                {format_instructions}

                Only output the json instance, anything else will result in a failed generation.
            """,
        )

        prompts: list[str] = [
            prompt_template.format(
                judge_prompt=judge_prompt,
                format_instructions=build_format_instructions(dynamic_model),
            )
            for judge_prompt in judge_prompts
        ]

        unparsed_responses: list[str] = cast(
            list[str],
            self.inference_engine(
                dataset=[
                    {"source": prompt, "data_classification_policy": ["public"]}
                    for prompt in prompts
                ]
            ),
        )

        on_failure_default_options = []
        if valid_outputs is not None:
            if isinstance(valid_outputs, list):
                on_failure_default_options = valid_outputs
            else:
                on_failure_default_options = list(
                    range(valid_outputs[0], valid_outputs[1] + 1, 1)
                )

        else:
            on_failure_default_options = [""]

        on_failure_default_options = on_failure_default_options * len(judge_prompts)
        parser = BatchRepairParser(
            inference_engine=self.inference_engine,
            max_retries=3,
            on_generation_failure=self.on_generation_failure,
        )

        parsed_responses, parsing_metadatas = parser.parse_and_repair(
            unparsed_responses=unparsed_responses,
            on_failure_default={selected_field: on_failure_default_options},
            model_classes=[dynamic_model] * len(unparsed_responses),
        )

        explanations: list[str] = [r.explanation for r in parsed_responses]  # type: ignore
        selected_options: list[str] = [
            getattr(r, "selected_option", str(getattr(r, "selected_score", None)))
            for r in parsed_responses
        ]

        return [
            DirectInstanceResult(
                selected_option=selected_option,
                score=float(selected_option) if is_float(selected_option) else None,
                explanation=explanation,
                metadata={
                    **parsing_metadata,
                    "prompt": prompt,
                    "unparsed_response": unparsed_response,
                },
            )
            for selected_option, explanation, prompt, unparsed_response, parsing_metadata in zip(
                selected_options,
                explanations,
                prompts,
                unparsed_responses,
                parsing_metadatas,
            )
        ]



import logging
import os
import uuid
from textwrap import dedent
from typing import cast

from evalassist.judges.batch_repair_parser import BatchRepairParser
from evalassist.judges.utils import (
    build_format_instructions,
    generate_dynamic_pydantic_model,
)
from evalassist.main import DirectInstanceDTO
from pydantic import Field
from unitxt.inference import InferenceEngine

from ..api_types import (
    CriteriaWithOptionsDTO,
    DirectActionTypeEnum,
    DirectAIActionRequest,
    DomainEnum,
    GenerationLengthEnum,
    PersonaEnum,
    TaskEnum,
)
from ..const import generation_length_to_sentence_count
from ..utils import to_snake_case

logger = logging.getLogger(__name__)


def get_data_path(task: TaskEnum, domain: DomainEnum):
    return os.path.join(
        os.path.dirname(os.path.abspath(__file__)),
        "data",
        to_snake_case(task.value),
        "source_data",
        "source.jsonl",
    )


class DirectActionGenerator:
    def __init__(
        self,
        action: DirectActionTypeEnum,
        prompt: str | None,
        inference_engine: InferenceEngine,
    ):
        self.action = action
        self.prompt = prompt
        self.inference_engine = inference_engine
        # intialize model

        self.action_third_person_dict = {
            DirectActionTypeEnum.REPHRASE: "rephrases",
            DirectActionTypeEnum.SHORTER: "shortens",
            DirectActionTypeEnum.LONGER: "elaborates on",
        }

        self.action_infinitive_person_dict = {
            DirectActionTypeEnum.REPHRASE: "to rephrase",
            DirectActionTypeEnum.SHORTER: "to shorten",
            DirectActionTypeEnum.LONGER: "to elaborate on",
        }

        self.action_past_dict = {
            DirectActionTypeEnum.REPHRASE: "rephrased",
            DirectActionTypeEnum.SHORTER: "shortened",
            DirectActionTypeEnum.LONGER: "elaborated",
        }

        self.parser = BatchRepairParser(
            inference_engine=self.inference_engine,
            max_retries=3,
            on_generation_failure="random",
        )

    def generate(self, direct_ai_action: DirectAIActionRequest):
        if self.action == DirectActionTypeEnum.CUSTOM:
            action_tag = "<custom_action>"
            dynamic_model = generate_dynamic_pydantic_model(
                model_name="structured_output_model",
                field_definitions=[
                    (
                        "response",
                        str,
                        Field(
                            ...,
                            description="the selection to apply the action to",
                        ),
                        [],
                    ),
                ],
            )
            format_instructions = build_format_instructions(dynamic_model)
            text_with_selection = direct_ai_action.text.replace(
                direct_ai_action.selection,
                action_tag + direct_ai_action.selection + action_tag,
            )
            # prompt templates
            system_prompt_template = dedent(
                """\
                You will be provided with:

                - A selected text

                - A text containing that selection, with the selection marked using {action_tag} tags

                Your task is to {action_description} the selected text such that:

                - It preserves the original meaning and intent

                - It fits seamlessly into the original text, both semantically and grammatically

                ✅ The generated selection must not disrupt the sentence structure or introduce grammatical errors (e.g., missing prepositions or incorrect tense).
                🚫 Do not introduce any new information that is not present in the original text.

                Selection:
                {selection}

                Text with selection (wrapped in-between {action_tag} tags):
                {text_with_selection}

                {format_instructions}
                Don't forget to enclose the response value in double quotes.
            """,
            )

            system_prompt = system_prompt_template.format(
                text_with_selection=text_with_selection,
                selection=direct_ai_action.selection,
                format_instructions=format_instructions,
                action_description=self.prompt,
                action_tag=action_tag,
            )
        elif self.action == DirectActionTypeEnum.REGENERATE:
            dynamic_model = generate_dynamic_pydantic_model(
                model_name="structured_output_model",
                field_definitions=[
                    (
                        "response",
                        str,
                        Field(
                            ...,
                            description="the selection to regenerate",
                        ),
                        [],
                    ),
                ],
            )
            format_instructions = build_format_instructions(dynamic_model)
            action_str = direct_ai_action.action.value.lower()
            action_tag = "<regenerate>"
            text_with_selection = direct_ai_action.text.replace(
                direct_ai_action.selection,
                action_tag + direct_ai_action.selection + action_tag,
            )
            # prompt templates
            system_prompt_template = dedent("""\
                You will be provided with:
                - A selected text
                - A text containing that selection, with the selection marked using <regenerate> tags
                - Your task is to substitute the selected text with a counterfactual example to diversify perspective, demographic, or approach. It should fit seamlessly into the original text. The regenerated selection must not disrupt the sentence structure or introduce grammatical errors (e.g., missing prepositions or incorrect tense).
                - Examples: “toddler” changed to “adult”, “terrorist” changed to “diplomat”, “men” changed to “women”, “easy” changed to “difficult”, “great” changed to “poor”

                Selection:
                {selection}

                Text with selection (wrapped in-between <regenerate> tags):
                {text_with_selection}

                {format_instructions}
                Don't forget to enclose the response value in double quotes.
                """)
            system_prompt = system_prompt_template.format(
                text_with_selection=text_with_selection,
                selection=direct_ai_action.selection,
                format_instructions=format_instructions,
            )
        else:
            dynamic_model = generate_dynamic_pydantic_model(
                model_name="structured_output_model",
                field_definitions=[
                    (
                        "response",
                        str,
                        Field(
                            ...,
                            description=f"the selection to {self.action.value.lower()}",
                        ),
                        [],
                    ),
                ],
            )
            format_instructions = build_format_instructions(dynamic_model)
            action_str = direct_ai_action.action.value.lower()
            action_tag = f"<{action_str}>"
            text_with_selection = direct_ai_action.text.replace(
                direct_ai_action.selection,
                action_tag + direct_ai_action.selection + action_tag,
            )
            # prompt templates
            system_prompt_template = dedent("""\
                You will be provided with:

                - A selected text

                - A text containing that selection, with the selection marked using {action_tag} tags

                Your task is {action_infinitive} the selected text such that:

                - It preserves the original meaning and intent

                - It fits seamlessly into the original text, both semantically and grammatically

                ✅ The {action_past} selection must not disrupt the sentence structure or introduce grammatical errors (e.g., missing prepositions or incorrect tense).
                🚫 Do not introduce any new information that is not present in the original text.

                - If the selection is equal to the whole text, your task is {action_infinitive} the whole text.
                - Examples: “toddler” changed to “kid”, “terrorist” changed to “extremist”, “men” changed to “human”, “easy” changed to “simple”, “great” changed to “excellent”

                Selection:
                {selection}

                Text with selection (wrapped in-between {action_tag} tags):
                {text_with_selection}

                {format_instructions}
                Don't forget to enclose the response value in double quotes.
                """)

            system_prompt = system_prompt_template.format(
                text_with_selection=text_with_selection,
                selection=direct_ai_action.selection,
                action_third_person=self.action_third_person_dict[self.action],
                action_infinitive=self.action_infinitive_person_dict[self.action],
                action_past=self.action_past_dict[self.action],
                action_tag=action_tag,
                format_instructions=format_instructions,
            )

        prompt = system_prompt

        logger.debug(f"Direct AI action prompt:\n{prompt}")

        unparsed_response: str = cast(
            str, self.inference_engine.infer([{"source": prompt}])[0]
        )
        logger.debug(f"Direct AI action unparsed response:\n{unparsed_response}")

        parsed_responses, metadata = self.parser.parse_and_repair(
            [unparsed_response], [dynamic_model]
        )
        parsed_response = parsed_responses[0]
        return parsed_response.model_dump()["response"]


class Generator:
    def __init__(
        self,
        inference_engine: InferenceEngine,
        criteria: CriteriaWithOptionsDTO,
        generation_length: GenerationLengthEnum | None,
        task: TaskEnum | None,
        domain: DomainEnum | None,
        persona: PersonaEnum | None,
        per_criteria_option_count: dict[str, int],
        borderline_count: int,
    ):
        self.inference_engine = inference_engine
        self.criteria: CriteriaWithOptionsDTO = criteria
        self.generation_length = generation_length
        self.task = task
        self.domain = domain
        self.persona = persona
        self.per_criteria_option_count = per_criteria_option_count
        self.borderline_count = borderline_count
        self.has_context_variables = (
            len(
                self.criteria.context_fields
                if self.criteria.context_fields is not None
                else []
            )
            > 0
        )

        self.parser = BatchRepairParser(
            inference_engine=self.inference_engine,
            max_retries=3,
            on_generation_failure="raise",
        )

        if self.task == TaskEnum.QUESTION_ANSWERING:
            self.dynamic_model = generate_dynamic_pydantic_model(
                model_name="structured_output_model",
                field_definitions=[
                    (
                        self.criteria.to_evaluate_field,
                        str,
                        Field(
                            ...,
                            description="the answer to the question",
                        ),
                        [],
                    ),
                ],
            )
            self.format_instructions = build_format_instructions(self.dynamic_model)

            # prompt templates
            self.system_prompt_template = dedent("""\
                You will be asked to generate an answer to a question according to the following requirements:

                Criteria name: {dimension}
                Criteria description: {dimension_description}
                Criteria dimension target: {target}
                {target_description_section}

                Your task is to generate an answer that STRICTLY follows this requirement. This is for evaluation purposes.

                Important:
                {domain_section}{persona_section}{generation_length_section}- Focus exclusively on the specified dimension and target
                - Make sure your answer clearly demonstrates the described characteristics
                - Do not mention the criteria in your answer - Simply generate an answer to the question that embodies the characteristics
                """)

            self.query_template = "Please generate an answer to the following question:\n\n{context_section}\n\n{format_instructions}"

        elif self.task == TaskEnum.SUMMARIZATION:
            self.dynamic_model = generate_dynamic_pydantic_model(
                model_name="structured_output_model",
                field_definitions=[
                    (
                        self.criteria.context_fields[0],
                        str,
                        Field(
                            ...,
                            description=f"the {self.criteria.context_fields[0]}'s summary",
                        ),
                        [],
                    ),
                ],
            )
            self.format_instructions = build_format_instructions(self.dynamic_model)

            # prompt templates
            self.system_prompt_template = dedent("""\
                You will be given some source text and will be asked to generate a summary according to a specific target criteria.

                You should generate a summary that matches the following requirements:
                Criteria name: {dimension}
                Criteria description: {dimension_description}
                Criteria dimension target: {target}
                {target_description_section}

                Your task is to generate a summary that STRICTLY follows this requirement. This is for evaluation purposes.

                Important:
                {domain_section}{persona_section}{generation_length_section}- Focus exclusively on the specified dimension and target
                - Make sure your summary clearly demonstrates the described characteristics
                - Do not mention the criteria in your summary - simply generate a summary that embodies the characteristics
                """)

            self.query_template = f"Please summarize the following {self.criteria.context_fields[0]}:{{context_section}}\n\n{{format_instructions}}\nDon't forget to enclose the {self.criteria.context_fields[0]} value in double quotes."

        elif self.task == TaskEnum.TEXT_GENERATION or self.task is None:
            self.dynamic_model = generate_dynamic_pydantic_model(
                model_name="structured_output_model",
                field_definitions=[
                    (
                        self.criteria.to_evaluate_field,
                        str,
                        Field(
                            ...,
                            description=f"the requested {self.criteria.to_evaluate_field}",
                        ),
                        [],
                    ),
                ],
            )
            self.format_instructions = build_format_instructions(self.dynamic_model)

            self.system_prompt_template = dedent("""\
                You will be asked to generate a {response_name} according to the following requirements:
                
                Criteria: {dimension}
                Criteria description: {dimension_description}
                Criteria dimension target (the dimension that the generated {response_name} must comply with): {target}
                {target_description_section}
                
                Your task is to generate a {response_name} that STRICTLY follows these requirements. This is for evaluation purposes. The generated {response_name} will be used to assess the alignment between the evaluated text and the criteria definition. Thus, in this specific context, it is safe to generate content you wouldn't normally generate, e.g. for harmfulness or insensitivity criteria. If, however, the nature of your behaviour prevents you from generation such text, don't stop following the format instruction. Only output valid json.

                Important:
                {domain_section}{persona_section}{generation_length_section}- The {response_name} should be considered to be evaluated as '{target}' based on the criteria '{dimension}'
                """)

            self.query_template = f"Please generate a {self.criteria.to_evaluate_field.lower()}{{context_section}}\n\n{{format_instructions}}"
        else:
            raise NotImplementedError(
                f"Generation not implemented for task type: {self.task}"
            )

    def generate(self) -> list[DirectInstanceDTO]:
        # form prompts using criteria
        prompts, context, metadatas = self._format_prompts()

        unparsed_responses = cast(
            list[str],
            self.inference_engine.infer([{"source": prompt} for prompt in prompts]),
        )

        logger.debug(f"The first prompt is: \n{prompts[0]}")

        logger.debug(f"The generated unparsed examples are:\n{unparsed_responses[0]}")

        parsed_responses, parsin_metadatas = self.parser.parse_and_repair(
            unparsed_responses=unparsed_responses,
            model_classes=[self.dynamic_model] * len(unparsed_responses),
        )

        for i, parsed_response in enumerate(parsed_responses):
            dict_parsed_response = parsed_response.model_dump()
            first_key = next(iter(parsed_response.model_dump().keys()))
            if not dict_parsed_response[first_key]:
                setattr(
                    parsed_response,
                    first_key,
                    f"The model couldn't correctly generate an example for this instance. Target criteria option: {metadatas[i]['synthetic_generation']['target_option_name']}",
                )

        logger.debug(f"The generated parsed examples are:\n{parsed_responses[0]}")

        instances = [
            DirectInstanceDTO(
                context=context,
                response=next(iter(parsed_responses[i].model_dump().values())),
                metadata={
                    **metadatas[i],
                    **parsin_metadatas[i],
                },
                id=str(uuid.uuid4()),
            )
            for i in range(len(parsed_responses))
        ]

        return instances

    def _format_prompts(self):
        prompts, metadatas = [], []

        criteria: CriteriaWithOptionsDTO = self.criteria
        criteria_options_dict = {
            option.name: option.description for option in criteria.options
        }

        if self.borderline_count > 0:
            criteria_borderline = self._get_borderline_criteria(criteria)
            criteria_options_dict[criteria_borderline["name"]] = criteria_borderline[
                "description"
            ]
            # Replace the borderline count by the synthetically generated borderline
            self.per_criteria_option_count[criteria_borderline["name"]] = (
                self.borderline_count
            )
        if self.domain is not None:
            domain_section = f"- The generated {self.criteria.to_evaluate_field.lower()} is going to be evaluated on the {self.domain.value} domain\n"
        else:
            domain_section = ""

        if self.persona is not None:
            persona_section = f"- Adopt the following persona: {self.persona.lower()}\n"
        else:
            persona_section = ""

        if self.generation_length is not None:
            generation_length_section = f"- The generated {self.criteria.to_evaluate_field.lower()}'s length should be {self.generation_length.value.lower()} ({generation_length_to_sentence_count[self.generation_length]} long).\n"
        else:
            generation_length_section = ""

        context: dict[str, str] = {}
        if self.has_context_variables:
            context = self._generate_synthetic_context()

            if (
                self.task == TaskEnum.SUMMARIZATION
                or self.task == TaskEnum.QUESTION_ANSWERING
            ):
                context_section = f"\n{context[self.criteria.context_fields[0]]}"
            else:
                context_placeholders = "\n".join(
                    [
                        f"{name}: {context[name]}"
                        for name in self.criteria.context_fields
                    ]
                )
                context_section = (
                    f" based on the following context:\n\n{context_placeholders}"
                )
        else:
            context_section = ""
        for criteria_option_name in self.per_criteria_option_count.keys():
            criteria_option_description = criteria_options_dict[criteria_option_name]
            if criteria_option_description:
                target_description_section = (
                    f"Criteria dimension description: {criteria_option_description}"
                )
            else:
                target_description_section = ""

            system_prompt_params = {
                "dimension": self.criteria.name,
                "dimension_description": self.criteria.description,
                "target": criteria_option_name,
                "target_description_section": target_description_section,
                "response_name": self.criteria.to_evaluate_field.lower(),
                "domain_section": domain_section,
                "persona_section": persona_section,
                "generation_length_section": generation_length_section,
            }

            system_prompt = self.system_prompt_template.format(**system_prompt_params)

            query = self.query_template.format(
                context_section=context_section,
                format_instructions=self.format_instructions,
            )

            prompt = system_prompt + "\n\n" + query
            prompts.extend(
                [prompt] * self.per_criteria_option_count[criteria_option_name]
            )
            metadata = {
                "synthetic_generation": {
                    "model_name": self.inference_engine.get_engine_id(),
                    "criteria_name": self.criteria.name,
                    "target_option_name": criteria_option_name,
                    "target_option_description": criteria_option_description,
                    "prompt": prompt,
                    "data_length": self.generation_length.value
                    if self.generation_length
                    else None,
                    "task": self.task.value if self.task else None,
                    "domain": self.domain.value if self.domain else None,
                    "persona": self.persona.value if self.persona else None,
                }
            }
            metadatas.extend(
                [metadata] * self.per_criteria_option_count[criteria_option_name]
            )

        return prompts, context, metadatas

    def _generate_synthetic_context(self) -> dict[str, str]:
        if self.task == TaskEnum.SUMMARIZATION:
            system_prompt_template = dedent("""\
                Your task is to generate a sample paragraph considering the following information:

                - The generated text is intended to be used to generate a summary.
                - The generated text should be 10-20 sentences long.
                {domain_section}
                """)
            # domain_section = f"- Pick a topic that is under the {self.domain.value} domain. The generated text should be related to this topic." if self.domain is not None else ""
            domain_section = (
                f"- You are working within the {self.domain.value} domain. "
                f"First, pick a specific topic _within_ {self.domain.value} (don't pick the domain itself as a topic) "
                if self.domain is not None
                else ""
            )
            system_prompt = system_prompt_template.format(
                domain_section=domain_section,
            )
            dynamic_model = generate_dynamic_pydantic_model(
                model_name="structured_output_model",
                field_definitions=[
                    (
                        "text",
                        str,
                        Field(
                            ...,
                            description="the text to generate",
                        ),
                        [],
                    ),
                ],
            )
            format_instructions = build_format_instructions(self.dynamic_model)
        else:
            system_prompt_template = dedent("""\
                You will be provided with a list of context variable names. Your task is to generate example values that simulate *real-life cases* for each of these context variables, considering the following information:

                - Context variables to generate: {context_names}.
                - The generated context is intended to be used to generate a '{response_name}'.{task_section}{domain_section}{persona_section}
                - For each context variable, generate realistic, natural examples that could be used in a dataset for this evaluation.
                - Do not describe the variable itself — instead, produce the actual kind of text or data it would contain in real life.
                - The generated context should make sense as input for generating a {response_name} that will later be evaluated using the {criteria_name} criterion (“{criteria_description}”).
                - If a variable name is vague (e.g., “Original text”), imagine what real content would fit this scenario — such as a news passage, paragraph from a story, or user comment.
                - Be concise but natural; summaries may span several sentences, while questions may be short.
                - Creativity is encouraged, but keep the examples plausible and contextually meaningful.""")
            task_section = (
                f"\n- The generated context is part of a dataset that conforms to a {self.task.value} task.\n"
                if self.task is not None
                else ""
            )
            domain_section = (
                f"- The generated context should be related to the {self.domain.value} domain.\n"
                if self.domain is not None
                else ""
            )
            persona_section = (
                f"- The generated context will be used by the following persona: {self.persona.lower()}.\n"
                if self.persona is not None
                else ""
            )
            system_prompt = system_prompt_template.format(
                context_names=", ".join(f"'{s}'" for s in self.criteria.context_fields),
                criteria_name=self.criteria.name,
                criteria_description=self.criteria.description,
                response_name=self.criteria.to_evaluate_field,
                task_section=task_section,
                domain_section=domain_section,
                persona_section=persona_section,
            )
            dynamic_model = generate_dynamic_pydantic_model(
                model_name="structured_output_model",
                field_definitions=[
                    (
                        context_name,
                        str,
                        Field(
                            ...,
                            description=f"The {context_name} to generate",
                        ),
                        [],
                    )
                    for context_name in self.criteria.context_fields
                ],
            )

        format_instructions = build_format_instructions(dynamic_model)

        query_template = dedent(
            f"""\
                You must output only valid JSON with no extra text.
                Use the following schema and formatting rules:
                {format_instructions}
                """
        )

        query = query_template

        prompt = system_prompt + query
        unparsed_response = cast(
            str, self.inference_engine.infer([{"source": prompt}])[0]
        )

        logger.debug(f"The prompt used for synthetic generation is:\n{prompt}")
        logger.debug(f"The synthetic generation response is:\n{unparsed_response}")

        parsed_responses, metadatas = self.parser.parse_and_repair(
            unparsed_responses=[unparsed_response], model_classes=[dynamic_model]
        )

        return parsed_responses[0].model_dump()

    def _get_borderline_criteria(self, criteria: CriteriaWithOptionsDTO):
        criteria_options = criteria.options
        if len(criteria_options) < 2:
            raise ValueError(
                "Need to specify at least two criteria to generate borderline case."
            )

        dynamic_model = generate_dynamic_pydantic_model(
            model_name="structured_output_model",
            field_definitions=[
                (
                    "name",
                    str,
                    Field(
                        ...,
                        description="the name of borderline criteria",
                    ),
                    [],
                ),
                (
                    "description",
                    str,
                    Field(
                        ...,
                        description="the description of borderline criteria",
                    ),
                    [],
                ),
            ],
        )
        criteria_format_instructions = build_format_instructions(dynamic_model)

        criteria_options_list = [
            f"{option.name}: {option.description}" for option in criteria_options
        ]
        criteria_options_section = "\n".join(criteria_options_list)

        query = f"You will be provided with a criteria. The criteria is composed by a name, a description and a set of criteria options. Describe a borderline criteria option that lies between the criteria options\n\nCriteria name: {criteria.name}\nCriteria description: {criteria.description}\nCriteria options:\n{criteria_options_section}\n\nProvide a natural language description of what it means to be a borderline case among these criteria options. Your description should mirror the style and format of the original criteria options but describe the subtle ways in which the case partially satisfies multiple criteria while not fully satisfying any single one.\n\n{criteria_format_instructions}"

        logger.debug(f"The borderline criteria generation prompt is \n{query}")

        borderline_criteria_unparsed = cast(
            str, self.inference_engine.infer([{"source": query}])[0]
        )
        logger.debug(
            f"The unparsed borderline criteria is:\n{borderline_criteria_unparsed}"
        )
        borderline_criteria_parsed, metadata = self.parser.parse_and_repair(
            unparsed_responses=[borderline_criteria_unparsed],
            model_classes=[dynamic_model],
        )

        return borderline_criteria_parsed[0].model_dump()



from typing import cast

from evalassist.judges.utils import get_to_evaluate_text

from .base import BaseDirectJudge, BasePairwiseJudge
from .types import (
    Criteria,
    DirectInstanceResult,
    DirectPositionalBiasResult,
    Instance,
    PairwiseInstanceResult,
    SingleSystemPairwiseInstanceResult,
)


class DummyDirectJudge(BaseDirectJudge):
    def get_name(self) -> str:
        return "dummy"

    def _run(
        self,
        instances: list[Instance],
        criteria: list[Criteria],
    ) -> list[DirectInstanceResult]:
        return [
            DirectInstanceResult(
                instance=instances[0],
                criteria=criteria[0],
                selected_option=criteria[0].options[0].name,
                explanation="explanation",
                positional_bias=DirectPositionalBiasResult(
                    detected=False,
                ),
            )
            for _ in range(len(instances))
        ]


class DummyPairwiseJudge(BasePairwiseJudge):
    def get_name(self) -> str:
        return "dummy"

    def _run(
        self,
        instances: list[Instance],
        criteria: list[Criteria],
    ) -> list[PairwiseInstanceResult]:
        results: list[PairwiseInstanceResult] = []
        systems_per_instance = len(
            cast(list[str], get_to_evaluate_text(instances[0], criteria[0]))
        )
        comparisons_per_instance = systems_per_instance - 1
        for i, instance in enumerate(instances):
            instance_result: list[SingleSystemPairwiseInstanceResult] = []
            instance_result.append(
                SingleSystemPairwiseInstanceResult(
                    contest_results=[True for _ in range(comparisons_per_instance)],
                    compared_to=[True for _ in range(comparisons_per_instance)],
                    explanations=[
                        "Explanations" for _ in range(comparisons_per_instance)
                    ],
                    positional_bias=[False for _ in range(comparisons_per_instance)],
                    winrate=1.0,
                    ranking=1,
                    selections=[1 for _ in range(comparisons_per_instance)],
                )
            )
            results.append(
                PairwiseInstanceResult(
                    selected_option="0",
                    per_system_results=instance_result,
                )
            )
        return results



import logging
import logging.config
import os
import pathlib

from evalassist.const import LOG_DIR


def configure_logging():
    handlers = ["console"]
    dict_config: dict = {
        "version": 1,
        "disable_existing_loggers": False,
        "formatters": {
            "standard": {
                "format": "%(asctime)s - %(name)s - %(lineno)d - %(levelname)s - %(message)s"
            }
        },
        "handlers": {
            "console": {
                "class": "logging.StreamHandler",
                "level": "INFO",  # or configurable via env / CLI flag
                "formatter": "standard",
                "stream": "ext://sys.stdout",
            },
        },
        "root": {"handlers": handlers, "level": "DEBUG", "propagate": False},
        # 👇 Add this
        "loggers": {
            "httpx": {
                "level": "WARNING",
                "handlers": [],
                "propagate": True,
            },
            "LiteLLM": {
                "level": "WARNING",
                "handlers": [],
                "propagate": True,
            },
            # Sometimes LiteLLM uses lowercase logger names:
            "litellm": {
                "level": "WARNING",
                "handlers": [],
                "propagate": True,
            },
        },
    }

    if LOG_DIR:
        pathlib.Path(LOG_DIR).mkdir(parents=True, exist_ok=True)
        handlers.append("file")
        dict_config["handlers"]["file"] = {
            "class": "logging.FileHandler",
            "level": "DEBUG",
            "formatter": "standard",
            "filename": os.path.join(LOG_DIR, "evalassist.log"),
            "encoding": "utf-8",
            "mode": "a",
        }

    logging.config.dictConfig(dict_config)
